#include <limits>
#include <iostream>
#include <iomanip>
using namespace std;


/*  nbits    a,     0
, nbits <= 64. */
void print_bits(unsigned long long a, int nbits)
{
    unsigned long long b = a;
    int bits[64];
    for (int i = 0; i < nbits; ++i)
    {
        bits[i] = b & 1; /* b % 2 */
        b >>= 1; /* b /= 2 */
    }
    for (int i = nbits - 1; i >= 0; --i)
        cout << bits[i];
}


/*  nbits    a,    , 
 , nbits <= 64. */
void print_fraction_hex(unsigned long long a, int nbits)
{
    /*       ,  4
       (  ..   ) */
    int multiplier = 1 << ((4 - (nbits % 4)) % 4);
    unsigned long long b = a * multiplier;
    int nhexs = (nbits + 3) / 4;
    int hexs[16];
    for (int i = 0; i < nhexs; ++i)
    {
        hexs[i] = b & 15; /* b % 16 */
        b >>= 4; /* b /= 16 */
    }
    cout << hex;
    for (int i = nhexs - 1; i >= 0; --i)
        cout << hexs[i];
    cout << dec;
}


/*      ,
   x -   , w -    ,
   t -    . */
void print_fp(void * x, int w, int t)
{
    /*     x     x_as_int */
    unsigned x_as_int = *((unsigned *)x);
    /*  : x_as_int = *reinterpret_cast<unsigned*>(x); */

    /*   ,   */
    unsigned sign_bit = x_as_int >> (w + t);

    /*   w ,   */
    unsigned exponent_bits = (x_as_int >> t) & ((1 << w) - 1);

    /*   t ,   */
    unsigned significand_bits = x_as_int & ((1 << t) - 1);

    /*     3 : ,  ,   */
    cout << (sign_bit ? 1 : 0) << " ";
    print_bits(exponent_bits, w);
    cout << " ";
    print_bits(significand_bits, t);
    cout << endl;

    /*     NaN */
    if (exponent_bits == (1 << w) - 1)
    {
        /*    */
        if (significand_bits == 0)
            if (sign_bit)
                cout << "- reserved infinity" << endl;
            else
                cout << "+ reserved infinity" << endl;
        else
            /*    NaN */
            cout << "  reserved NaN" << endl;
        return;
    }

    /*  */
    int emax = (1 << (w - 1)) - 1;
    int emin = 1 - emax;
    int bias = emax;

    /*   */
    if (sign_bit)
        cout << "- ";
    else
        cout << "+ ";

    /*   0    */
    if (exponent_bits == 0)
    {
        /*   0 */
        if (significand_bits == 0)
            cout << setw(w - 5) << "[" << emin << "] 0" << endl;
        else
        {
            /*     ,   0 */
            cout << setw(w - 5) << "[" << emin << "] 0,";
            print_fraction_hex(significand_bits, t);
            cout << endl;
        }
        return;
    }

    /*       ,     ,
      1 */
    cout << setw(w) << (int)exponent_bits - bias;
    cout << " 1,";
    print_fraction_hex(significand_bits, t);
    cout << endl;
}


/*        . */
void print_float(float x)
{
    /*   float: 8    ,
    23     */
    print_fp((void*)&x, 8, 23);
}


int main()
{
    /*    0,125   float */
    cout << 0.125f << " = " << endl;
    print_float(0.125f);
    cout << endl;

    /*   +0   float */
    cout << "positive zero = " << 0.0f << " = " << endl;
    print_float(0.0f);
    cout << endl;

    /*   -0   float */
    cout << "negative zero = " << -0.0f << " = " << endl;
    print_float(-0.0f);
    cout << endl;

    /*       
     float */
    cout << "min positive subnormal float value ~ "
         << numeric_limits<float>::denorm_min() << " = "<< endl;
    print_float(numeric_limits<float>::denorm_min());
    cout << endl;

    /*       
     float */
    cout << "min positive normal float value ~ "
         << numeric_limits<float>::min() << " = "<< endl;
    print_float(numeric_limits<float>::min());
    cout << endl;

    /*        float */
    cout << "max normal float value ~ "
         << numeric_limits<float>::max() << " = "<< endl;
    print_float(numeric_limits<float>::max());
    cout << endl;

    /*     (  1   
      )   float */
    cout << "machine epsilon ~ "
         << numeric_limits<float>::epsilon() << " = "<< endl;
    print_float(numeric_limits<float>::epsilon());
    cout << endl;

    /*   +   float */
    cout << "positive infinity = " << numeric_limits<float>::infinity()
         << " = " << endl;
    print_float(numeric_limits<float>::infinity());
    cout << endl;

    /*   -   float */
    cout << "negative infinity = " << -numeric_limits<float>::infinity()
         << " = " << endl;
    print_float(-numeric_limits<float>::infinity());
    cout << endl;

    /*   quite NaN   float */
    cout << "quiet nan = " << endl;
    print_float(numeric_limits<float>::quiet_NaN());
    cout << endl;

    return 0;
}
