#include <iostream>
#include <iomanip>
#include <math.h>
using namespace std;



/*  nbits    a,     0
, nbits <= 64. */
void print_bits(unsigned long long a, int nbits)
{
    unsigned long long b = a;
    int bits[64];
    for (int i = 0; i < nbits; ++i)
    {
        bits[i] = b & 1; /* b % 2 */
        b >>= 1; /* b /= 2 */
    }
    for (int i = nbits - 1; i >= 0; --i)
        cout << bits[i];
}


/*  nbits    a,    , 
 , nbits <= 64. */
void print_fraction_hex(unsigned long long a, int nbits)
{
    /*       ,  4
       (  ..   ) */
    int multiplier = 1 << ((4 - (nbits % 4)) % 4);
    unsigned long long b = a * multiplier;
    int nhexs = (nbits + 3) / 4;
    int hexs[16];
    for (int i = 0; i < nhexs; ++i)
    {
        hexs[i] = b & 15; /* b % 16 */
        b >>= 4; /* b /= 16 */
    }
    cout << hex;
    for (int i = nhexs - 1; i >= 0; --i)
        cout << hexs[i];
    cout << dec;
}


/*      ,
   x -   , w -    ,
   t -    . */
void print_fp(void * x, int w, int t)
{
    /*     x     x_as_int */
    unsigned x_as_int = *((unsigned *)x);
    /*  : x_as_int = *reinterpret_cast<unsigned*>(x); */

    /*   ,   */
    unsigned sign_bit = x_as_int >> (w + t);

    /*   w ,   */
    unsigned exponent_bits = (x_as_int >> t) & ((1 << w) - 1);

    /*   t ,   */
    unsigned significand_bits = x_as_int & ((1 << t) - 1);

    /*     3 : ,  ,   */
    cout << (sign_bit ? 1 : 0) << " ";
    print_bits(exponent_bits, w);
    cout << " ";
    print_bits(significand_bits, t);
    cout << endl;

    /*     NaN */
    if (exponent_bits == (1 << w) - 1)
    {
        /*    */
        if (significand_bits == 0)
            if (sign_bit)
                cout << "- reserved infinity" << endl;
            else
                cout << "+ reserved infinity" << endl;
        else
            /*    NaN */
            cout << "  reserved NaN" << endl;
        return;
    }

    /*  */
    int emax = (1 << (w - 1)) - 1;
    int emin = 1 - emax;
    int bias = emax;

    /*   */
    if (sign_bit)
        cout << "- ";
    else
        cout << "+ ";

    /*   0    */
    if (exponent_bits == 0)
    {
        /*   0 */
        if (significand_bits == 0)
            cout << setw(w - 5) << "[" << emin << "] 0" << endl;
        else
        {
            /*     ,   0 */
            cout << setw(w - 5) << "[" << emin << "] 0,";
            print_fraction_hex(significand_bits, t);
            cout << endl;
        }
        return;
    }

    /*       ,     ,
      1 */
    cout << setw(w) << (int)exponent_bits - bias;
    cout << " 1,";
    print_fraction_hex(significand_bits, t);
    cout << endl;
}


/*        . */
void print_float(float x)
{
    /*   float: 8    ,
    23     */
    print_fp((void*)&x, 8, 23);
}


//    ax^2 + bx + c = 0
//   ; ,  a    0.
template <typename FP_TYPE>
void solveNaive(FP_TYPE a, FP_TYPE b, FP_TYPE c, FP_TYPE & root1,
    FP_TYPE & root2)
{

    FP_TYPE delta = b * b - 4 * a * c; // 
    if (delta >= 0)
    {
        //   
        FP_TYPE t = sqrt(delta);
        root1 = (-b - t) / (2 * a);
        root2 = (-b + t) / (2 * a);
        //      
        cout << "   -b - t = " << -b - t << endl;
        cout << "   -b + t = " << -b + t << endl;
    }
}


const float eps = 1e-4f;

//   sqrt(1 + x) - 1     
//   ,     |x| < eps.
float sqrtf1(float x)
{
    //  sqrt(1 + x)       ,
    // 1 ,    :
    // C(1/2, i) * x^i,  i  1  +, 
    // (1/2, i) -    1/2  i.
    int numSteps = 10; //  ,   x  
    float sum = 0;
    float coeff = 0.5f; //    
    float arg = x; // x   
    // 
    for (int j = 1; j <= numSteps; ++j)
    {
        //   
        sum += coeff * arg;

        //     
        coeff *= (0.5f - j) / (j + 1);
        arg *= x;
    }

    return sum;
}

//    ax^2 + bx + c = 0   
//        ;
// ,  a    0.
void solveSeriesSumFloat(float a, float b, float c, float & root1,
    float & root2)
{
    float delta = b * b - 4 * a * c; // 
    if (delta >= 0)
    {
        float t = sqrtf(delta);

        //     :
        // root1 = (-b - t) / (2 * a);
        // root2 = (-b + t) / (2 * a).
        //   t / b = sign(b) * sqrt(1 - 4 * a * c / (b * b)),
        //  arg = - 4 * a * c / (b * b)   0,    
        //  b      root1  root2 
        //   0       
        //    .      
        //   sqrt(1 - 4 * a * c / (b * b)) - 1.
        float arg = - 4 * a * c / (b * b);
        if (fabs(arg) < eps)
        {
            //  b > 0,        root2,  
            //     ,  root1   ,
            //     root1     , 
            // root2   .
            if (b > 0)
            {
                root1 = (-b - t) / (2 * a);
                // root2 = (-b + t) / (2 * a) =
                //    = (b / (2 * a)) * (sqrt(1 + arg) - 1)
                root2 = (b / (2 * a)) * sqrtf1(arg);
            }
            else
            {
                // root1 = (-b - t) / (2 * a) =
                //    = (b / (2 * a)) * (sqrt(1 + arg) - 1)
                // (   , .. sign(b) = -1)
                root1 = (b / (2 * a)) * sqrtf1(arg);
                root2 = (-b + t) / (2 * a);
            }
        }
        else
        {
            root1 = (-b - t) / (2 * a);
            root2 = (-b + t) / (2 * a);
        }
    }
}


//    ax^2 + bx + c = 0   
//       ;
// ,  a    0.
void solveAlgebraicTransformFloat(float a, float b, float c, float & root1,
    float & root2)
{
    float delta = b * b - 4 * a * c; // 
    if (delta >= 0)
    {
        float t = sqrtf(delta);
        
        //     :
        // root1 = (-b - t) / (2 * a);
        // root2 = (-b + t) / (2 * a).
        //   ,  
        // (  )  (    ):
        // -b - t = (b^2 - t^2) / (-b + t) = 4 * a * c / (-b + t) =>
        // root1 = (-b - t) / (2 * a) = 2 * c / (-b + t);
        // -b + t = (b^2 - t^2) / (-b - t) = 4 * a * c / (-b - t) =>
        // root2 = (-b + t) / (2 * a) = 2 * c / (-b - t).
        //  |-b - t| > |-b + t|,    root1  
        // ,  root2  ,    root2 
        // ,  root1  .
        if (fabsf(-b - t) > fabsf(-b + t))
        {
            root1 = (-b - t) / (2 * a);
            root2 = 2 * c / (-b - t);
        }
        else
        {
            root1 = 2 * c / (-b + t);
            root2 = (-b + t) / (2 * a);
        }
    }
}


int main()
{
    //   9      
    cout << setprecision(9);

    //   
    // ax^2 + bx + c = 0,
    //     float  double 
    float a = 0.25f;
    float b = 1000.0f;
    float c = 1.0f;

    //   
    cout << "Exact roots:" << endl;
    //  ,   double
    double root1_exact = -3999.998999999749999;
    double root2_exact = -0.00100000025000078;
    cout << "   root1 = " << root1_exact << endl;
    cout << "   root2 = " << root2_exact << endl;
    cout << endl;

    // ""     
    cout << "Naive version double:" << endl;
    double root1_double, root2_double;
    solveNaive<double>(a, b, c, root1_double, root2_double);
    cout << "   root1 = " << root1_double << endl;
    cout << "   root2 = " << root2_double << endl << endl;

    //      
    cout << "Naive version float:" << endl;
    float root1_float, root2_float;
    solveNaive<float>(a, b, c, root1_float, root2_float);
    cout << "   root1 = " << root1_float << endl;
    cout << "   root2 = " << root2_float << endl << endl;

    //         
    cout << "Series sum version float:" << endl;
    float root1_series, root2_series;
    solveSeriesSumFloat(a, b, c, root1_series, root2_series);
    cout << "   root1 = " << root1_series << endl;
    cout << "   root2 = " << root2_series << endl << endl;

    //       
    // 
    cout << "Algebraic transform version float:" << endl;
    float root1_algebraic, root2_algebraic;
    solveAlgebraicTransformFloat(a, b, c, root1_algebraic, root2_algebraic);
    cout << "   root1 = " << root1_algebraic << endl;
    cout << "   root2 = " << root2_algebraic << endl;

    return 0;
}
